/*
    UDPEQ
    UDP load balancing over parallel routes
    Copyright (C) 2001 Truxton Fulton <trux@truxton.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>

extern void srand48() ;
extern long lrand48() ;

/* these defines should become command line parameters */
#define DEBUG 8
#define NUM_HASH_BUCKETS 8192
#define STRICT_ORDERING 0  /* do we need to assemble packets in order? (NIY) */
#define UDP_SEND_SOCKBUFSIZE 65536
#define IDLE_TIMEOUT_SEC 32  /* if a channel goes dry for 32 seconds, consider it inactive */
#define RECONNECT_TIMEOUT_SEC 48  /* if a channel goes dry for 48 seconds, try to reconnect */
#define MIN_SEND_PERIOD_USEC 2800000  /* send a packet at least every 2.8 seconds */
#define STATUS_PERIOD_USEC 28000000  /* print status every 28 seconds */
#define KEEPALIVE_PERIOD_USEC 4000000  /* send keepalive every 4 seconds */
#define MAX_LATENCY_USEC 2871000  /* never utilize a channel with more than 2.8 secs additional latency */
#define UDPEQ_PRIORITY -20  /* need high priority to avoid delaying packets */
#define TOLERABLE_PROBABILITY_OF_LOSS 0.05  /* we want at least 95% of our packets to reach their destination */
#define HISTORY_QUEUE_LENGTH 1024  /* max num of timestamps in the history queue */
#define PACKET_LOSS_WINDOW_USEC 64000000  /* look at num sent and received over last minute (64 seconds) */
#define MIN_PROBABILITY_OF_ARRIVAL 0.05  /* with less than 5% probability of arrival, dont consider the latency estimate accurate */

/* these defines are immutable */
#define STRLEN 8192
#define BUFLEN 8192
#define UDPEQ_HEADER_SIZE ((8*sizeof(long))+2)
#define UDPEQ_DATA_MAGIC 0xE29078BE

struct timeval last_send_time ;
long global_sequence_counter=0 ;
long hash_table_seq[NUM_HASH_BUCKETS] ;
long hash_table_rnd[NUM_HASH_BUCKETS] ;

enum udpeq_channel_type
{
  LI,  /* local input */
  LO,  /* local output */
  RI,  /* remote input */
  RO,  /* remote output */
  UNKNOWN
} ;

struct udpeq_channel
{
  enum udpeq_channel_type type ;
  char *hostname ;
  int port ;
  int local_port ;
  /* a master/slave pair share the same fd */
  /* master_channel and slave_channel cannot both be non-NULL */
  struct udpeq_channel *master_channel ;  /* in the case of this being a passive reply channel, else NULL */
  struct udpeq_channel *slave_channel ;  /* reverse link, NULL if replies on this channel are ignored */
  char slave_reply_address[128] ;
  int fd ;
  struct sockaddr_in sa ;
  int active ;
  char symbol ;
  char remote_symbols[128] ;
  int send_flag ;
  int recv_flag ;
  struct timeval time_of_recent_activity ;
  struct timeval time_of_recent_send ;
  struct timeval send_time_history[HISTORY_QUEUE_LENGTH] ;
  unsigned long send_latency_history[HISTORY_QUEUE_LENGTH] ;
  int send_time_history_length ;
  int send_time_history_index ;
  struct timeval receive_time_history[HISTORY_QUEUE_LENGTH] ;
  int receive_time_history_length ;
  int receive_time_history_index ;
  unsigned long inactive_time ;
  long latency ;  /* estimate of current latency (usec) */
  long aggregate_latency ;  /* aggregate recent latency (usec) */
  long average_latency ;  /* average recent latency (usec) */
  long display_latency ;  /* latency normalized with min_latency at 0 */
  double score ;
  unsigned long num_sends_in_last_minute ;
  unsigned long num_received_in_last_minute ;
  double estimated_probability_of_arrival ;

  struct udpeq_channel *previous ;
} ;

char *name_of_type(struct udpeq_channel *c)
{
  char *result=NULL ;
  switch(c->type)
    {
    case LI :
      result="LI     " ;
      break ;;;
    case LO :
      result="LO     " ;
      break ;;;
    case RI :
      if(c->master_channel!=NULL)
	result="RI(psv)" ;
      else
	if(c->slave_channel!=NULL)
	  result="RIO    " ;
	else
	  result="RI     " ;
      break ;;;
    case RO :
      if(c->master_channel!=NULL)
	result="RO(psv)" ;
      else
	if(c->slave_channel!=NULL)
	  result="ROI    " ;
	else
	  result="RO     " ;
      break ;;;
    default :
      result="UNKNOWN" ;
      break ;;;
    }
  return(result) ;
}

void usage(char *program_name)
{
  fprintf(stderr,"
---------------- USAGE: ----------------
%s -LI host:port -LO host:port [-RI host:port -RO host:port]...
You must specify exactly one LI and one LO option.
You must specify at least one RI and one RO option.
LI means local input (talking to local cipe process).
LO means local output (talking to local cipe process).
RI means remote input (talking to remote udpeq process).
RO means remote output (talking to remote udpeq process).
RIO means remote input with reply output.
ROI means remote output with reply input.
udpeq listens on LI for packets and encapsulates them
over an optimized equalized tunnel to a remote udpeq.
The encapsulated packets may travel over any of the
RO sockets.  Return encapsulated packets are gathered
at the RI sockets, unencapsulated, and output to LO.
LI and LO may be the same port, similarly for R[IO].
The host name may be omitted to indicate the local host.
\n",program_name) ;
}

struct udpeq_channel *parse_command_line(int argc, char **argv)
{
  struct udpeq_channel *result=NULL,*c,*master_channel ;
  enum udpeq_channel_type type, reply_channel ;
  int li=0,lo=0,ri=0,ro=0,fatal=0 ;
  int local_port ;
  struct timeval now ;
  char symbol='A' ;
  char *colon ;
  int i,j,k ;

  gettimeofday(&now,NULL) ;
  memcpy(&last_send_time,&now,sizeof(struct timeval)) ;  /* init */
  for(i=1;i<argc;i++)
    {
      reply_channel=UNKNOWN ;
      if(!strcmp(argv[i],"-LI"))
	{
	  type=LI ;
	  li++ ;
	}
      if(!strcmp(argv[i],"-LO"))
	{
	  type=LO ;
	  lo++ ;
	}
      if(!strcmp(argv[i],"-RI"))
	{
	  type=RI ;
	  ri++ ;
	}
      if(!strcmp(argv[i],"-RO"))
	{
	  type=RO ;
	  ro++ ;
	}
      if(!strcmp(argv[i],"-RIO"))
	{
	  type=RI ;
	  reply_channel = RO ;
	  ri++ ;
	  ro++ ;
	}
      if(!strcmp(argv[i],"-ROI"))
	{
	  type=RO ;
	  reply_channel = RI ;
	  ro++ ;
	  ri++ ;
	}
      if(type==UNKNOWN)
	{
	  fprintf(stderr,"%s ERROR: unknown option \"%s\"\n",argv[0],argv[i]) ;
	  fatal=1 ;
	}
      else
	{
	  i++ ;
	  if(i==argc)
	    {
	      fprintf(stderr,"%s ERROR: %s missing socket specification\n",argv[0],argv[i-1]) ;
	      fatal=1 ;
	    }
	  else
	    {
	      colon=strstr(argv[i],":") ;
	      if(colon==NULL)
		{
		  fprintf(stderr,"%s ERROR: invalid socket specification \"%s\"\n",argv[0],argv[i]) ;
		  fatal=1 ;
		}
	      else
		{
		  *colon=0 ;
		  master_channel=NULL ;
		  for(;type!=UNKNOWN;)
		    {
		      c=(struct udpeq_channel *)malloc(sizeof(struct udpeq_channel)) ;
		      c->type=type ;
		      c->hostname=strdup(argv[i]) ;
		      if(strlen(c->hostname)<1)
			{
			  free(c->hostname) ;
			  if((type==LI)||(type==RI))
			    c->hostname=strdup("0.0.0.0") ;
			  else  /* output socket */
			    c->hostname=strdup("127.0.0.1") ;
			}
		      c->port=atoi(colon+1) ;
		      c->local_port=c->port ;
		      c->master_channel=master_channel ;
		      if(master_channel!=NULL)
			master_channel->slave_channel=c ;
		      c->fd= -1 ;
		      c->active=0 ;
		      c->send_time_history_length=0 ;
		      c->send_time_history_index=0 ;
		      c->receive_time_history_length=0 ;
		      c->receive_time_history_index=0 ;
		      memcpy(&(c->time_of_recent_send),&now,sizeof(struct timeval)) ;  /* init */
		      memcpy(&(c->time_of_recent_activity),&now,sizeof(struct timeval)) ;  /* init */
		      c->previous=result ;
		      c->symbol=symbol++ ;
		      c->remote_symbols[0]=0 ;
		      c->inactive_time=0 ;
		      c->num_sends_in_last_minute=0 ;
		      c->num_received_in_last_minute=0 ;
		      c->estimated_probability_of_arrival=0.0 ;

		      c->latency=0 ;
		      c->average_latency=0 ;
		      c->aggregate_latency=0 ;

		      result=c ;
		      type=reply_channel ;
		      if(type!=UNKNOWN)
			{
			  reply_channel=UNKNOWN ;
			  master_channel=c ;
			}
		    }
		}
	    }
	}
    }

  if(li!=1)
    {
      fprintf(stderr,"%s: ERROR: you must specify exactly one -LI option\n",argv[0]) ;
      fatal=1 ;
    }
  if(lo!=1)
    {
      fprintf(stderr,"%s: ERROR: you must specify exactly one -LO option\n",argv[0]) ;
      fatal=1 ;
    }
  if(ri<1)
    {
      fprintf(stderr,"%s: ERROR: you must specify at least one -RI option\n",argv[0]) ;
      fatal=1 ;
    }
  if(ro<1)
    {
      fprintf(stderr,"%s: ERROR: you must specify at least one -RO option\n",argv[0]) ;
      fatal=1 ;
    }

  if(fatal)
    result=NULL ;

  /* packets sent to LO should have return address of LI */
  for(c=result;c!=NULL;c=c->previous)
    if(c->type==LI)
      local_port=c->port ;
  for(c=result;c!=NULL;c=c->previous)
    if(c->type==LO)
      c->local_port=local_port ;

  return(result) ;
}

void close_channel(struct udpeq_channel *c)
{
  int err ;

  err=0 ;

  if(c==NULL)
    err=1 ;

  if(!err)
    {
      if(c->fd>=0)
	close(c->fd) ;
      c->fd= -1 ;
      c->active=0 ;
      if(c->master_channel!=NULL)
	{
	  if(DEBUG>4)
	    fprintf(stderr,"\nclosing master channel %c : %s (%s:%d)  ",
		    c->symbol,name_of_type(c),c->hostname,c->port) ;
	  c->master_channel->fd= -1 ;
	}

      if(c->slave_channel!=NULL)
	{
	  if(DEBUG>4)
	    fprintf(stderr,"\nclosing slave channel %c : %s (%s:%d)  ",
		    c->symbol,name_of_type(c),c->hostname,c->port) ;
	  c->slave_channel->fd= -1 ;
	}
    }
}

void open_channel(struct udpeq_channel *c)
{
  int err, opt ;
  char hostip[STRLEN] ;
  char message[STRLEN] ;
  struct hostent *hostaddr ;
  unsigned long int addrval ;

  err=0 ;

  if(c==NULL)
    err=1 ;

  if(!err)
    {
      if(c->master_channel!=NULL)
	{
	  if(c->master_channel->fd<0)
	    open_channel(c->master_channel) ;
	  c->fd=c->master_channel->fd ;
	}
      else
	{
	  if(DEBUG>3)
	    fprintf(stderr,"\nopening %c : %s socket (%s:%d)...  ",
		    c->symbol,name_of_type(c),c->hostname,c->port) ;
	  c->fd=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP) ;
	  if(c->fd<0)
	    {
	      perror("socket") ;
	      err=1 ;
	    }
	  if(!err)
	    {
	      opt = 1 ;
	      if(setsockopt(c->fd, SOL_SOCKET, SO_REUSEADDR,(char *) &opt, sizeof(opt))==-1)
		{
		  perror("setsockopt (1)") ;
		  close(c->fd) ;
		  c->fd= -1 ;
		  err=1 ;
		}
	    }
	  if(!err)
	    {
	      opt = UDP_SEND_SOCKBUFSIZE ;
	      if(setsockopt(c->fd, SOL_SOCKET, SO_SNDBUF,(char *) &opt, sizeof(opt))==-1)
		{
		  perror("setsockopt (1)") ;
		  close(c->fd) ;
		  c->fd= -1 ;
		  err=1 ;
		}
	    }
	  
	  if(!err)
	    {
	      hostip[0]=0 ;
	      
	      if((addrval = inet_addr(c->hostname)) != -1)  /* user gave us an IP address */
		{
		  sprintf(hostip,"%s",c->hostname) ;
		}
	      else  /* resolve hostname */
		{
		  hostaddr=gethostbyname(c->hostname) ;
		  if(hostaddr!=NULL)
		    {
		      sprintf(hostip,"%d.%d.%d.%d",
			      (((unsigned char *) (hostaddr->h_addr_list[0]))[0]),
			      (((unsigned char *) (hostaddr->h_addr_list[0]))[1]),
			      (((unsigned char *) (hostaddr->h_addr_list[0]))[2]),
			      (((unsigned char *) (hostaddr->h_addr_list[0]))[3])) ;
		    }
		  else
		    {
		      int ipb0, ipb1, ipb2, ipb3 ;
		      if(4==sscanf(c->hostname,"%d.%d.%d.%d",&ipb0,&ipb1,&ipb2,&ipb3))
			{
			  sprintf(hostip,"%d.%d.%d.%d",ipb0,ipb1,ipb2,ipb3) ;
			  printf("open_channel: cannot resolve host name, trying literal IP address %s\n",hostip) ;
			}
		    }
		  if(hostip[0]==0)
		    {
		      fprintf(stderr,"open_channel: cannot resolve host name : %s\n",c->hostname) ;
		      close(c->fd) ;
		      c->fd= -1 ;
		      err=1 ;
		    }
		  else
		    if((addrval = inet_addr(hostip)) == -1)
		      {
			fprintf(stderr,"open_channel: invalid host IP : %s\n",c->hostname) ;
			close(c->fd) ;
			c->fd= -1 ;
			err=1 ;
		      }
		}
	    }
	}
    }
      
  if(!err)
    {
      if((c->type==LI)||
	 /*	 (c->type==LO) || */
	 ((c->type==RI) && (c->master_channel==NULL)))
	{
	  c->sa.sin_family = AF_INET ;
	  c->sa.sin_addr.s_addr = addrval ;
	  c->sa.sin_port = htons(c->local_port) ;
	  if(bind(c->fd,(struct sockaddr *) &c->sa,sizeof(struct sockaddr_in))<0)
	    {
	      if(DEBUG>0)
		{
		  fprintf(stderr,"error binding %s socket (%s:%d) fd==%d\n",
			  name_of_type(c),c->hostname,c->port,c->fd) ;
		  perror("open_channel: bind") ;
		}
	      close_channel(c) ;
	      err=1 ;
	    }
	}
      if((c->type==LO) ||
	 ((c->type==RO) && (c->master_channel==NULL)))
	{
	  c->sa.sin_family = AF_INET ;
	  c->sa.sin_addr.s_addr = addrval ;
	  c->sa.sin_port = htons(c->port) ;
	  if(connect(c->fd,(struct sockaddr*) &c->sa,sizeof(struct sockaddr_in))<0)
	    {
	      if(DEBUG>0)
		{
		  fprintf(stderr,"error connecting %s socket (%s:%d) fd==%d\n",
			  name_of_type(c),c->hostname,c->port,c->fd) ;
		  perror("open_channel: connect") ;
		}
	      close_channel(c) ;
	      err=1 ;
	    }
	}
    }
  
  if(!err)
    if(c->master_channel==NULL)
      {
	if((opt=fcntl(c->fd, F_GETFL))<0)
	  {
	    perror("F_GETFL") ;
	    err=1 ;
	  }
	else if (fcntl(c->fd, F_SETFL, opt|O_NONBLOCK)<0)
	  {
	    perror("F_SETFL") ;
	    err=1 ;
	  }
	if(err)
	  {
	    fprintf(stderr,"ERROR: cannot set nonblocking mode\n") ;
	    close(c->fd) ;
	    c->fd= -1 ;
	  }
      }

  if(!err)
    {
      c->active=1 ;
      gettimeofday(&(c->time_of_recent_activity),NULL) ;  /* open */
      if(DEBUG>6)
	fprintf(stderr,"successfully opened %s socket (%s:%d)\n",
		name_of_type(c),c->hostname,c->port) ;
    }

  if(!err)
    if(c->slave_channel!=NULL)
      open_channel(c->slave_channel) ;
}

int disburse_data(struct udpeq_channel *channel_tail,char *buf, int buflen, struct timeval *now)
{
  struct udpeq_channel *c ;
  int numsends=0 ;
  int i,n,r,length ;
  unsigned long delta ;
  struct timeval *then ;

  for(c=channel_tail;c!=NULL;c=c->previous)
    if((c->send_flag) && (c->fd>=0))
      {
	c->send_flag=0 ;
	length=buflen ;
	if(DEBUG>8)
	  fprintf(stderr,"sending %d bytes to %s channel (%s:%d) active=%d...\n",
		  length,name_of_type(c),c->hostname,c->port,c->active) ;
	if(c->type==RO)
	  buf[length++]=c->symbol ;
	/* r=sendto(c->fd,buf,buflen,0,&(c->sa),sizeof(struct sockaddr_in)) ; */
	r=send(c->fd,buf,length,0) ;
	if(r<0)
	  {
	    if(DEBUG>6)
	      fprintf(stderr,"error %d sending %d bytes to %c : %s channel (%s:%d) fd==%d\n",
		      errno,length,c->symbol,name_of_type(c),c->hostname,c->port,c->fd) ;
	    if(c->active)
	      {
		perror("send") ;
		fprintf(stderr,"errno==%d\n",errno) ;
		switch(errno)
		  {
		  case ENOTCONN :
		    if(DEBUG>5)
		      fprintf(stderr,"\nmarking channel %c inactive  ",c->symbol) ;
		    c->active=0 ;
		    break ;;;
		  }
	      }
	  }
	else
	  if(r!=length)
	    {
	      if(DEBUG>0)
		fprintf(stderr,"ERROR: only sent %d of %d bytes.\n",r,length) ;
	    }
	  else
	    {
	      n=c->send_time_history_length ;
	      i=c->send_time_history_index ;
	      i-=n ;
	      if(i<0)
		i+=HISTORY_QUEUE_LENGTH ;

	      for(;n>0;n--)
		{
		  then=&(c->send_time_history[i]) ;
		  delta=((1000000*(now->tv_sec-then->tv_sec)) +
			 (now->tv_usec-then->tv_usec)) ;
		  if(delta>PACKET_LOSS_WINDOW_USEC)
		    {
		      c->aggregate_latency-=c->send_latency_history[i] ;
		      c->send_time_history_length-- ;
		    }
		  else
		    n=0 ;
		  i=(i+1)%HISTORY_QUEUE_LENGTH ;
		}

	      numsends++ ;
	      c->send_flag=1 ;
	      memcpy(&(c->time_of_recent_send),now,sizeof(struct timeval)) ;
	      memcpy(&last_send_time,now,sizeof(struct timeval)) ;

	      i=c->send_time_history_index ;

	      if(c->send_time_history_length==HISTORY_QUEUE_LENGTH)
		c->aggregate_latency-=c->send_latency_history[i] ;

	      memcpy(&(c->send_time_history[i]),now,sizeof(struct timeval)) ;
	      c->send_latency_history[i]=c->latency ;
	      c->aggregate_latency+=c->latency ;

	      c->send_time_history_index++ ;
	      c->send_time_history_index%=HISTORY_QUEUE_LENGTH ;
	      if(c->send_time_history_length<HISTORY_QUEUE_LENGTH)
		c->send_time_history_length++ ;
	      c->num_sends_in_last_minute=c->send_time_history_length ;

	      c->average_latency=c->aggregate_latency / ((long) c->num_sends_in_last_minute) ;
	    }
      }
  return(numsends) ;
}

void print_status(struct udpeq_channel *channel_tail)
{
  struct udpeq_channel *c ;
  long min_latency ;
  int i ;

  for(c=channel_tail;c!=NULL;c=c->previous)
    c->display_latency=c->latency ;

  if(0)  /* normalize to zero */
    {
      i=0 ;
      min_latency=0 ;
      for(c=channel_tail;c!=NULL;c=c->previous)
	if((c->active)&&(c->type==RO))
	  {
	    if((i==0)||(min_latency>c->display_latency))
	      min_latency=c->display_latency ;
	    i++ ;
	  }
      for(c=channel_tail;c!=NULL;c=c->previous)
	if((c->active)&&(c->type==RO))
	  c->display_latency-=min_latency ;
      
      i=0 ;
      min_latency=0 ;
      for(c=channel_tail;c!=NULL;c=c->previous)
	if((c->active)&&(c->type==RI))
	  {
	    if((i==0)||(min_latency>c->display_latency))
	      min_latency=c->display_latency ;
	    i++ ;
	  }
      for(c=channel_tail;c!=NULL;c=c->previous)
	if((c->active)&&(c->type==RI))
	  c->display_latency-=min_latency ;
    }

  for(c=channel_tail;c!=NULL;c=c->previous)
    {
      fprintf(stderr,"\n  %c (%2s) %s : ",c->symbol,c->remote_symbols,name_of_type(c)) ;
      if(!c->active)
	fprintf(stderr,"inactive (%ld)",c->inactive_time/1000000) ;
      else
	{
	  fprintf(stderr,"L=%0.3f AL=%0.3f",((double) c->display_latency)/1000000.0,((double) c->average_latency)/1000000.0) ;
	  if(c->type==RO)
	    fprintf(stderr," P=%0.3f (%ld of %ld)",c->estimated_probability_of_arrival,c->num_received_in_last_minute,c->num_sends_in_last_minute) ;
	}
      fprintf(stderr,"  (%15s:%5d) {fd=%2d}",c->hostname,c->port,c->fd) ;

      /* erode probability numbers */
      /*
	need to do this better 

      if(c->num_received_in_last_minute>0)
	c->num_received_in_last_minute/=2 ;
      if(c->num_sends_in_last_minute>0)
	c->num_sends_in_last_minute/=2 ;
      */
    }
  fprintf(stderr,"\n") ;
}

struct udpeq_channel **sorted_channels=NULL ;
int num_channels=0 ;

int channel_comparitor(const void *param_a, const void *param_b)
{
  int result ;
  struct udpeq_channel *a, *b ;

  a=*((struct udpeq_channel **) param_a) ;
  b=*((struct udpeq_channel **) param_b) ;

  if(a->score < b->score)
    result= 1 ;
  else if(a->score > b->score)
    result= -1 ;
  else
    result=0 ;

  return(result) ;
}

void decide_which_channels_to_use(struct udpeq_channel *channel_tail, struct timeval *now)
{
  struct udpeq_channel *c ;
  long delta, latency, lowest_latency, highest_latency ;
  int i, n, num_sends, min_num_sends, num_viable_channels ;
  double anticipated_probability_of_loss ;

  anticipated_probability_of_loss=1.0 ;
  num_viable_channels=0 ;
  highest_latency=MAX_LATENCY_USEC ;
  lowest_latency=0.0 ;
  num_sends=0 ;
  n=0 ;

  if(DEBUG>7)
    fprintf(stderr,"**(") ;

  for(c=channel_tail;c!=NULL;c=c->previous)
    {
      if((c->type==RO) && (c->active))
        {
          n++ ;
          if(n>num_channels)
            {
              if(DEBUG>0)
                fprintf(stderr,"\nrealloc sorting array to %d channels.  ",n) ;
              sorted_channels=(struct udpeq_channel **) realloc(sorted_channels,n*sizeof(struct udpeq_channel *)) ;
              num_channels=n ;
            }
          sorted_channels[n-1]=c ;
          latency=c->latency ;

          if((latency>highest_latency) || (n==1))
            highest_latency=latency ;   
          
          if(c->estimated_probability_of_arrival>=MIN_PROBABILITY_OF_ARRIVAL)
            {
              if((latency<lowest_latency) || (num_viable_channels==0))
                lowest_latency=latency ;   
           
              num_viable_channels++ ;
            }
          
          delta=((1000000*(now->tv_sec-c->time_of_recent_send.tv_sec)) +
                 (now->tv_usec-c->time_of_recent_send.tv_usec)) ;
          
          if(delta>MIN_SEND_PERIOD_USEC)
            {
              num_sends++ ;
              c->send_flag=2 ;
              anticipated_probability_of_loss*=(1.0 - c->estimated_probability_of_arrival) ;
            }
        }
    }


  if(DEBUG>7)
    fprintf(stderr,"%d,",num_sends) ;

  for(i=0;i<n;i++)
    {
      c=sorted_channels[i] ;
      if(c->estimated_probability_of_arrival<MIN_PROBABILITY_OF_ARRIVAL)
        latency=highest_latency ;
      else
        {
          latency=c->average_latency ;
          latency-=lowest_latency ;
        }
      c->score=0.0 - ((double) latency) ;
    }
  
  qsort(sorted_channels,n,sizeof(struct udpeq_channel *),channel_comparitor) ;

  /* always utilize the channel with best score */
  c=sorted_channels[0] ;
  if(!c->send_flag)
    {
      num_sends++ ;
      c->send_flag=1 ;
      anticipated_probability_of_loss*=(1.0 - c->estimated_probability_of_arrival) ;
    }

  if(DEBUG>7)
    fprintf(stderr,"%d,",num_sends) ;

  for(i=0;(i<n)&&(anticipated_probability_of_loss>TOLERABLE_PROBABILITY_OF_LOSS);i++)
    {
      c=sorted_channels[i] ;
      if(!c->send_flag)
	{
	  num_sends++ ;
	  c->send_flag=1 ;
	  anticipated_probability_of_loss*=(1.0 - c->estimated_probability_of_arrival) ;
	}
    }

  if(DEBUG>7)
    fprintf(stderr,"%d,",num_sends) ;

  for(c=channel_tail;c!=NULL;c=c->previous)
    if((c->type==RO) && (c->active))
      if((c->latency - lowest_latency) > MAX_LATENCY_USEC)
	if(c->send_flag)
	  {
	    c->send_flag-- ;
	    if(!(c->send_flag))
	      {
		if(DEBUG>6)
		  fprintf(stderr," <X %c (%ld) %ld X> ",c->symbol,c->latency,c->latency - lowest_latency) ;
		num_sends-- ;
	      }
	  }


  if(DEBUG>7)
    fprintf(stderr,"%d,",num_sends) ;

  min_num_sends=1 ;  /* always send at least one packet */

  if(num_sends<min_num_sends)
    if(DEBUG>4)
      fprintf(stderr,"WARNING! num_sends would have been %d < min %d\n",num_sends,min_num_sends) ;
  for(i=0;(num_sends<min_num_sends)&&(i<n);i++)
    {
      c=sorted_channels[i] ;
      if(!c->send_flag)
	{
	  c->send_flag=1 ;
	  num_sends++ ;
	}
    }

  if(DEBUG>7)
    fprintf(stderr,"%d)**",num_sends) ;

  if(num_sends==0)
    {
      if(DEBUG>0)
	fprintf(stderr,"\nERROR: no RO channels available. (n==%d) ",n) ;
    }
}

void update_estimated_probability(struct udpeq_channel *c)
{
  double ratio=0.0 ;

  if(c->num_sends_in_last_minute>0)
    ratio=((double) c->num_received_in_last_minute) / ((double) c->num_sends_in_last_minute) ;
  if(ratio>1.0)
    {
      /* fprintf(stderr,"WARNING: update_estimated_probability: rcvd %ld > sent %ld\n",c->num_received_in_last_minute,c->num_sends_in_last_minute) ; */
      ratio=1.0 ;
    }

  c->estimated_probability_of_arrival=ratio ;
}

void process_data(struct udpeq_channel *channel_tail,struct udpeq_channel *ic,char *buf,int buflen)
{
  struct udpeq_channel *c, *rc ;
  long magic, seq, rnd ;
  int i, hash ;
  int skipped_duplicate=0 ;
  int numsends ;
  struct timeval t0, t1, t2 ;
  double u0,u1 ;
  char s0,s1 ;
  int n0,n1 ;
  char rsymchar[2] ;
  char rc_char ;
  double dice ;
  int num_feedback_channels, slen ;
  long latency ;
  enum udpeq_channel_type ic_type ;
  unsigned long usecs_taken_to_send_data ;
  unsigned long num_received_in_last_minute=0 ;

  t0.tv_sec=0 ;
  t0.tv_usec=0 ;
  gettimeofday(&t1,NULL) ;

  for(c=channel_tail;c!=NULL;c=c->previous)
    c->send_flag=0 ;

  num_feedback_channels=0 ;
  for(c=channel_tail;c!=NULL;c=c->previous)
    if((c->type==RI)&&(c->active))
      num_feedback_channels++ ;

  rc=NULL ;
  dice=drand48()*((double) num_feedback_channels) ;
  for(c=channel_tail;(rc==NULL)&&(c!=NULL);c=c->previous)
    if((c->type==RI)&&(c->active))
      {
	dice-=1.0 ;
	if(dice<0.0)
	  rc=c ;
      }

  if(rc==NULL)
    {
      rc_char='?' ;
      latency=0 ;
    }
  else
    {
      slen=strlen(rc->remote_symbols) ;
      if(slen>0)
	rc_char=rc->remote_symbols[lrand48()%slen] ;
      else
	rc_char='?' ;
	
      latency=rc->latency ;
      num_received_in_last_minute=rc->num_received_in_last_minute ;
    }

  if(ic==NULL)  /* just a ping */
    {
      ic_type=LI ;
    }
  else
    {
      if(ic->type==RO)
	{
	  if(ic->master_channel!=NULL)
	    ic=ic->master_channel ;
	  else
	    if(ic->slave_channel!=NULL)
	      ic=ic->slave_channel ;
	}
      memcpy(&(ic->time_of_recent_activity),&t1,sizeof(struct timeval)) ;  /* recv */
      ic_type=ic->type ;
    }

  switch(ic_type)
    {
    case LI :
    case LO :  /* reply to LO should be treated as LI */
      /* prepend header */
      buflen+=UDPEQ_HEADER_SIZE ;
      buf-=UDPEQ_HEADER_SIZE ;
      seq=global_sequence_counter++ ;
      rnd=lrand48() ;
      *((long *) (buf+(0*sizeof(long))))=UDPEQ_DATA_MAGIC ;
      *((unsigned long *) (buf+(1*sizeof(unsigned long))))=t1.tv_sec ;
      *((unsigned long *) (buf+(2*sizeof(unsigned long))))=t1.tv_usec ;
      *((long *) (buf+(3*sizeof(long))))=seq ;
      *((long *) (buf+(4*sizeof(long))))=rnd ;
      *((long *) (buf+(5*sizeof(long))))=latency ;
      *((char *) (buf+(6*sizeof(long))))=rc_char ;
      *((unsigned long *) (buf+(7*sizeof(unsigned long))))=num_received_in_last_minute ;

      /* try to do load balancing on RO channels */
      decide_which_channels_to_use(channel_tail,&t1) ;
      break ;;;
    case RI :
      /* examine header */
      magic=*((long *) (buf+(0*sizeof(long)))) ;
      t0.tv_sec=*((unsigned long *) (buf+(1*sizeof(unsigned long)))) ;
      t0.tv_usec=*((unsigned long *) (buf+(2*sizeof(unsigned long)))) ;

      if(magic==UDPEQ_DATA_MAGIC)
	{
	  latency=((1000000*(t1.tv_sec-t0.tv_sec)) + (t1.tv_usec-t0.tv_usec)) ;
	  ic->latency=latency ;
	  buflen-- ;
	  sprintf(rsymchar,"%c",buf[buflen]) ;
	  if(strstr(ic->remote_symbols,rsymchar)==NULL)
	    strcat(ic->remote_symbols,rsymchar) ;
	  seq=*((unsigned long *) (buf+(3*sizeof(long)))) ;
	  rnd=*((unsigned long *) (buf+(4*sizeof(long)))) ;

	  hash=seq%NUM_HASH_BUCKETS ;
	  if((hash_table_seq[hash]==seq) && (hash_table_rnd[hash]==rnd))
	    {
	      /* skip duplicate packet */
	      skipped_duplicate=1 ;
	      if(DEBUG>8)
		fprintf(stderr,"skip duplicate packet %ld (rnd %lX)\n",seq,rnd) ;
	    }
	  else
	    {
	      hash_table_seq[hash]=seq ;
	      hash_table_rnd[hash]=rnd ;

	      /* continue examining the header for feedback info */
	      latency=*((long *) (buf+(5*sizeof(long)))) ;
	      rc_char=*((char *) (buf+(6*sizeof(long)))) ;
	      num_received_in_last_minute=*((unsigned long *) (buf+(7*sizeof(unsigned long)))) ;
	      if(DEBUG>7)
		fprintf(stderr,"<NRLM %ld (bl %d)>",num_received_in_last_minute,buflen) ;
	      if(DEBUG>7)
		fprintf(stderr,"{%c}",rc_char) ;
	      for(c=channel_tail;c!=NULL;c=c->previous)
		if(c->symbol==rc_char)
		  {
		    c->latency=latency ;
		    c->num_received_in_last_minute=num_received_in_last_minute ;
		    memcpy(&(c->time_of_recent_activity),&t1,sizeof(struct timeval)) ;  /* feedback */
		    update_estimated_probability(c) ;
		  }

	      /* we're done examining the header, strip it */
	      buflen-=UDPEQ_HEADER_SIZE ;
	      buf+=UDPEQ_HEADER_SIZE ;
	      if(buflen>0)
		for(c=channel_tail;c!=NULL;c=c->previous)
		  if(c->type==LO)
		    c->send_flag=1 ;
	    }	    
	}
      else
	{
	  fprintf(stderr,"bad magic %lX (seq %ld, rnd %lX)\n",magic,seq,rnd) ;
	}
      break ;;;
    default :
      fprintf(stderr,"ERROR: unexpected data (%d bytes) from %s channel (%s:%d)\n",
	      buflen,name_of_type(ic),ic->hostname,ic->port) ;
      break ;;;
    }

  numsends=disburse_data(channel_tail,buf,buflen,&t1) ;

  /* time taken to send data */
  gettimeofday(&t2,NULL) ;
  usecs_taken_to_send_data=((1000000*(t2.tv_sec-t1.tv_sec)) + (t2.tv_usec-t1.tv_usec)) ;

  if((ic_type==LI)||(ic_type==LO))  /* if sending to remote udpeq */
    {
      if(DEBUG>0)
	fprintf(stderr,"%d",numsends) ;
      if(DEBUG>0)
	for(c=channel_tail;c!=NULL;c=c->previous)
	  if(c->send_flag)
	    {
	      fprintf(stderr,"%c",c->symbol) ;
	      if(DEBUG>7)
		fprintf(stderr,":%0.3lf",c->score / 1000.0) ;
	    }
      if(DEBUG>8)
	fprintf(stderr,"{%ldus}",usecs_taken_to_send_data) ;
      if(DEBUG>0)
	fprintf(stderr,"[%04ld] ",seq%1024) ;
    }

  if(ic_type==RI)  /* if sending to local process */
    if(magic==UDPEQ_DATA_MAGIC)
      if(skipped_duplicate)
	{
	  if(DEBUG>0)
	    fprintf(stderr,"%c.",ic->symbol) ;
	}
      else
	{
	  if(DEBUG>0)
	    fprintf(stderr,"\n%c(%04ld) ",ic->symbol,(seq%1024)) ;
	}
}

void update_received_history(struct udpeq_channel *c, struct timeval *now)
{
  int i,n ;
  unsigned long delta ;
  struct timeval *then ;

  n=c->receive_time_history_length ;
  i=c->receive_time_history_index ;
  i-=n ;
  if(i<0)
    i+=HISTORY_QUEUE_LENGTH ;
  
  for(;n>0;n--)
    {
      then=&(c->receive_time_history[i]) ;
      delta=((1000000*(now->tv_sec-then->tv_sec)) +
	     (now->tv_usec-then->tv_usec)) ;
      if(delta>PACKET_LOSS_WINDOW_USEC)
	c->receive_time_history_length-- ;
      else
	n=0 ;
      i=(i+1)%HISTORY_QUEUE_LENGTH ;
    }
  
  memcpy(&(c->receive_time_history[c->receive_time_history_index]),now,sizeof(struct timeval)) ;
  c->receive_time_history_index++ ;
  c->receive_time_history_index%=HISTORY_QUEUE_LENGTH ;
  if(c->receive_time_history_length<HISTORY_QUEUE_LENGTH)
    c->receive_time_history_length++ ;
  c->num_received_in_last_minute=c->receive_time_history_length ;
}

void udpeq(struct udpeq_channel *channel_tail)
{
  struct udpeq_channel *c ;
  fd_set readfds ;
  struct timeval now, timeout ;
  struct timeval last_status_time ;
  unsigned long delta ;
  int r, selres, maxfd, buflen ;
  char buf[BUFLEN] ;
  char *offbuf ;
  char ipaddrs[128] ;
  struct sockaddr_in whofrom ;
  int fromlen,fromport ;

  gettimeofday(&last_status_time,NULL) ;

  /* print channel summary */
  for(c=channel_tail;c!=NULL;c=c->previous)
    fprintf(stderr,"%c : %s channel (%s:%d)\n",
	    c->symbol,name_of_type(c),c->hostname,c->port) ;

  for(;;)
    {
      /* open any sockets that are closed */
      /* !@#$ : this should be done asynchronously !!!  */
      for(c=channel_tail;c!=NULL;c=c->previous)
	if(c->fd<0)
	  open_channel(c) ;

      for(c=channel_tail;c!=NULL;c=c->previous)
	c->recv_flag=0 ;

      FD_ZERO(&readfds) ;
      maxfd=0 ;
      for(c=channel_tail;c!=NULL;c=c->previous)
	{
	  if((c->type==LI)||(c->type==LO)||(c->type==RI))
	    if(c->fd>=0)
	      {
		c->recv_flag=1 ;
		FD_SET(c->fd,&readfds) ;
		if(c->fd>maxfd)
		  maxfd=c->fd ;
	      }
	}
      timeout.tv_sec=4 ;
      timeout.tv_usec=0 ;
      selres=select(maxfd+1,&readfds,NULL,NULL,&timeout) ;
      
      gettimeofday(&now,NULL) ;

      if(selres<0)
	{
	  if(DEBUG>0)
	    perror("select") ;
	  if(errno==EBADF)
	    {
	      if(DEBUG>0)
		fprintf(stderr,"some channel is dead, let's find out which...\n") ;
	      for(c=channel_tail;c!=NULL;c=c->previous)
		{
		  FD_ZERO(&readfds) ;
		  FD_SET(c->fd,&readfds) ;
		  maxfd=c->fd ;
		  timeout.tv_sec=0 ;
		  timeout.tv_usec=0 ;
		  if(select(maxfd+1,&readfds,NULL,NULL,&timeout)== -1)
		    if(errno==EBADF)
		      {
			if(DEBUG>0)
			  fprintf(stderr,"discovered that %s socket (%s:%d) is dead\n",
				  name_of_type(c),c->hostname,c->port) ;
			close(c->fd) ;
			c->fd= -1 ;
		      }
		}
	    }	    
	}
      if(selres>0)
	for(c=channel_tail;c!=NULL;c=c->previous)
	  if(c->recv_flag)
	    if(FD_ISSET(c->fd,&readfds))
	      {
		offbuf=buf+UDPEQ_HEADER_SIZE ;
		buflen=BUFLEN-UDPEQ_HEADER_SIZE ;
		r=recvfrom(c->fd,offbuf,buflen,MSG_DONTWAIT|MSG_NOSIGNAL,&whofrom,&fromlen) ;
		if(r<0)
		  {
		    if(DEBUG>2)
		      fprintf(stderr,"error receiving packet from %s channel (%s:%d) fd==%d\n",
			      name_of_type(c),c->hostname,c->port,c->fd) ;
		    perror("recvfrom") ;
		    close_channel(c) ;
		  }
		if(r>0)
		  {
		    fromport=ntohs(whofrom.sin_port) ;
		    sprintf(ipaddrs,"%s",inet_ntoa(whofrom.sin_addr)) ;
		    if(DEBUG>8)
		      fprintf(stderr,"received %d bytes from %s:%d on %s %s:%d\n",
			      r,ipaddrs,fromport,name_of_type(c),c->hostname,c->port) ;
		    if(!c->active)
		      {
			if(DEBUG>5)
			  fprintf(stderr,"\nmarking channel %c active  ",c->symbol) ;
			c->active=1 ;
			memcpy(&(c->time_of_recent_activity),&now,sizeof(struct timeval)) ;  /* recv */
		      }
		    if(c->slave_channel!=NULL)
		      if(!c->slave_channel->active)
			{
			  c->slave_channel->sa.sin_family = AF_INET ;
			  memcpy(&(c->slave_channel->sa.sin_addr),
				 &(whofrom.sin_addr),
				 sizeof(whofrom.sin_addr)) ;
			  c->slave_channel->sa.sin_port = whofrom.sin_port ;
			  if(connect(c->slave_channel->fd,(struct sockaddr*) &c->slave_channel->sa,sizeof(struct sockaddr_in))<0)
			    {
			      sprintf(c->slave_reply_address,"<cannot reverse connect to %s:%d>",ipaddrs,fromport) ;
			      if(DEBUG>0)
				fprintf(stderr,"ERROR: cannot connect to remote %s:%d to send reply\n",
					ipaddrs,fromport) ;
			    }
			  else
			    {
			      sprintf(c->slave_reply_address,"%s:%d",ipaddrs,fromport) ;

			      if(DEBUG>5)
				fprintf(stderr,"\nmarking slave channel %c active  ",c->slave_channel->symbol) ;
			      c->slave_channel->active=1 ;
			      memcpy(&(c->slave_channel->time_of_recent_activity),&now,sizeof(struct timeval)) ;  /* recv on master */
			    }
			}
		    update_received_history(c,&now) ;
		    process_data(channel_tail,c,offbuf,r) ;
		  }
	      }

      gettimeofday(&now,NULL) ;

      for(c=channel_tail;c!=NULL;c=c->previous)
	if((c->type==RO)||(c->type==RI))
	  {
	    delta=(now.tv_sec-c->time_of_recent_activity.tv_sec) ;
	    c->inactive_time=delta ;
	    if(c->active)
	      {
		if(delta>IDLE_TIMEOUT_SEC)
		  {
		    if(DEBUG>4)
		      fprintf(stderr,"\nmarking idle (delta=%ld) %s channel %c (%s:%d) inactive  ",
			      delta,name_of_type(c),c->symbol,c->hostname,c->port) ;
		    c->active=0 ;
		  }
	      }
	    if(delta>RECONNECT_TIMEOUT_SEC)
	      {
		if(DEBUG>4)
		  fprintf(stderr,"\nattempt reconnect for idle (delta=%ld) %s channel %c (%s:%d)  ",
			  delta,name_of_type(c),c->symbol,c->hostname,c->port) ;
		close_channel(c) ;
	      }
	  }
      
      delta=((1000000*(now.tv_sec-last_status_time.tv_sec)) + (now.tv_usec-last_status_time.tv_usec)) ;
      if(delta>STATUS_PERIOD_USEC)
	{
	  memcpy(&last_status_time,&now,sizeof(struct timeval)) ;
	  print_status(channel_tail) ;
	}

      delta=((1000000*(now.tv_sec-last_send_time.tv_sec)) + (now.tv_usec-last_send_time.tv_usec)) ;
      if(delta>KEEPALIVE_PERIOD_USEC)
	{
	  for(c=channel_tail;c!=NULL;c=c->previous)
	    if((c->type==RO)&&(c->active))
	      c->send_flag=1 ;
	    else
	      c->send_flag=0 ;
	  offbuf=buf+UDPEQ_HEADER_SIZE ;
	  process_data(channel_tail,NULL,offbuf,0) ;
	  if(DEBUG>2)
	    fprintf(stderr,"!") ;
	}
    }
}

int main(int argc, char **argv)
{
  int result=0 ;
  struct udpeq_channel *channel_tail ;
  struct timeval now ;

  fprintf(stderr,"udpeq version %s\n",UDPEQ_VERSION) ;

  if(setpriority(PRIO_PROCESS, 0, UDPEQ_PRIORITY))
    {
      fprintf(stderr,"Cannot set process priority (need to be root)\n") ;
      fprintf(stderr,"Continuing anyway...\n") ;
      sleep(4) ;
    }

  gettimeofday(&now,NULL) ;
  srand48(((long) getpid())*now.tv_usec + (now.tv_sec^((long) &now))) ;
  channel_tail=parse_command_line(argc,argv) ;

  if(channel_tail!=NULL)
    udpeq(channel_tail) ;
  else
    {
      usage(argv[0]) ;
      result= -1 ;
    }

  return(result) ;
}
